Type: Package
Package: densratio
Title: Density Ratio Estimation
Version: 0.3.1
Authors@R: 
    person("Koji", "Makiyama", , "hoxo.smile@gmail.com", role = c("aut", "cre"))
Description: Density ratio estimation.  The estimated density ratio
    function can be used in many applications such as anomaly detection,
    change-point detection, covariate shift adaptation.  The implemented
    methods are uLSIF (Hido et al. (2011)
    <doi:10.1007/s10115-010-0283-2>), RuLSIF (Yamada et al. (2011)
    <doi:10.1162/NECO_a_00442>), and KLIEP (Sugiyama et al. (2007)
    <doi:10.1007/s10463-008-0197-x>).
License: MIT + file LICENSE
URL: https://github.com/hoxo-m/densratio
BugReports: https://github.com/hoxo-m/densratio/issues
Imports: utils
Suggests: graphics, knitr, mvtnorm, rmarkdown, stats, testthat
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-29 12:43:32 UTC; akagi
Author: Koji Makiyama [aut, cre]
Maintainer: Koji Makiyama <hoxo.smile@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-29 13:10:11 UTC
Built: R 4.6.0; ; 2026-05-01 20:12:05 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,KLIEP)
S3method(print,RuLSIF)
S3method(print,densratio)
S3method(print,uLSIF)
export(KLIEP)
export(RuLSIF)
export(densratio)
export(uLSIF)
importFrom(utils,str)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# An R Package for Density Ratio Estimation

#### *Koji MAKIYAMA (@hoxo-m)*

<!-- badges: start -->

[![CRAN
Version](https://www.r-pkg.org/badges/version/densratio)](https://CRAN.R-project.org/package=densratio)
[![CRAN
Downloads](https://cranlogs.r-pkg.org/badges/densratio)](https://CRAN.R-project.org/package=densratio)
[![R-CMD-check](https://github.com/hoxo-m/densratio/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/hoxo-m/densratio/actions/workflows/R-CMD-check.yaml)
[![Coverage
Status](https://coveralls.io/repos/github/hoxo-m/densratio/badge.svg?branch=master)](https://coveralls.io/github/hoxo-m/densratio?branch=master)
<!-- badges: end -->

## 1. Overview

**Density ratio estimation** is described as follows: for given two data
samples `x1` and `x2` from unknown distributions `p(x)` and `q(x)`
respectively, estimate `w(x) = p(x) / q(x)`, where `x1` and `x2` are
d-dimensional real numbers.

The estimated density ratio function `w(x)` can be used in many
applications such as **anomaly detection** \[Hido et al. 2011\],
**change-point detection** \[Liu et al. 2013\], and **covariate shift
adaptation** \[Sugiyama et al. 2007\]. Other useful applications about
density ratio estimation were summarized by \[Sugiyama et al. 2012\].

The package **densratio** provides a function `densratio()` that returns
an object with a method to estimate density ratio as
`compute_density_ratio()`.

For example,

``` r
set.seed(3)
x1 <- rnorm(200, mean = 1, sd = 1/8)
x2 <- rnorm(200, mean = 1, sd = 1/2)

library(densratio)
densratio_obj <- densratio(x1, x2)
```

The densratio object has a function `compute_density_ratio()` that can
compute the estimated density ratio `w_hat(x)` for any d-dimensional
input `x` (now d=1).

``` r
new_x <- seq(0, 2, by = 0.03)
w_hat <- densratio_obj$compute_density_ratio(new_x)

plot(new_x, w_hat, pch=19)
```

<img src="man/figures/README-compute-estimated-density-ratio-1.png" alt="" width="500" />

In this case, the true density ratio
`w(x) = p(x)/q(x) = Norm(1, 1/8) / Norm(1, 1/2)` is known. So we can
compare `w(x)` with the estimated density ratio `w-hat(x)`.

``` r
true_density_ratio <- function(x) dnorm(x, 1, 1/8) / dnorm(x, 1, 1/2)

plot(true_density_ratio, xlim=c(0, 2), lwd=2, col="red", xlab = "x", ylab = "Density Ratio")
plot(densratio_obj$compute_density_ratio, xlim=c(0, 2), lwd=2, col="green", add=TRUE)
legend("topright", legend=c(expression(w(x)), expression(hat(w)(x))), col=2:3, lty=1, lwd=2, pch=NA)
```

<img src="man/figures/README-compare-true-estimate-1.png" alt="" width="500" />

## 2. Installation

You can install the **densratio** package from
[CRAN](https://CRAN.R-project.org/package=densratio).

``` r
install.packages("densratio")
```

You can also install the package from
[GitHub](https://github.com/hoxo-m/densratio).

``` r
install.packages("remotes") # if you have not installed "remotes" package
remotes::install_github("hoxo-m/densratio")
```

The source code for **densratio** package is available on GitHub at

- <https://github.com/hoxo-m/densratio>.

## 3. Details

### 3.1 Basics

The package provides `densratio()`. The function returns an object that
has a function to compute estimated density ratio.

For data samples `x1` and `x2`,

``` r
set.seed(3)
x1 <- rnorm(200, mean = 1, sd = 1/8)
x2 <- rnorm(200, mean = 1, sd = 1/2)

library(densratio)
densratio_obj <- densratio(x1, x2)
```

In this case, `densratio_obj$compute_density_ratio()` can compute
estimated density ratio.

``` r
new_x <- seq(0, 2, by = 0.03)
w_hat <- densratio_obj$compute_density_ratio(new_x)

plot(new_x, w_hat, pch=19)
```

<img src="man/figures/README-basics-compute-estimated-density-ratio-1.png" alt="" width="500" />

### 3.2 Methods

`densratio()` has `method` argument that you can pass `"uLSIF"`,
`"RuLSIF"`, or `"KLIEP"`.

- **uLSIF** (unconstrained Least-Squares Importance Fitting) is the
  default method. This algorithm estimates density ratio by minimizing
  the squared loss. You can find more information in \[Kanamori et
  al. 2009\] and \[Hido et al. 2011\].
- **RuLSIF** (Relative unconstrained Least-Squares Importance Fitting).
  This algorithm estimates relative density ratio by minimizing the
  squared loss. You can find more information in \[Yamada et al. 2011\]
  and \[Liu et al. 2013\].
- **KLIEP** (Kullback-Leibler Importance Estimation Procedure). This
  algorithm estimates density ratio by minimizing Kullback-Leibler
  divergence. You can find more information in \[Sugiyama et al. 2007\].

The methods assume that density ratios are represented by linear model:

- `w(x) = theta_1 * K(x, c_1) + theta_2 * K(x, c_2) + ... + theta_b * K(x, c_b)`

where

- `K(x, c) = exp(-||x - c||^2 / (2 * sigma^2))`

is the Gaussian (RBF) kernel.

`densratio()` performs the following:

- Decides kernel parameter `sigma` by cross-validation,
- Optimizes the kernel weights `theta` (in other words, find the optimal
  coefficients of the linear model), and
- The parameters `sigma` and `theta` are saved into `densratio` object,
  and are used when to compute density ratio in the call
  `compute_density_ratio()`.

### 3.3 Result and Arguments

`densratio()` outputs the result like as follows:

    #> 
    #> Call:
    #> densratio(x1 = x1, x2 = x2, method = "uLSIF")
    #> 
    #> Kernel Information:
    #>   Kernel type:  Gaussian 
    #>   Number of kernels:  100 
    #>   Bandwidth (sigma):  0.1 
    #>   Centers:  num [1:100, 1] 0.907 1.093 1.18 1.136 1.046 ...
    #> 
    #> Kernel Weights:
    #>   num [1:100] 0.067455 0.040045 0.000459 0.016849 0.067084 ...
    #> 
    #> Regularization Parameter (lambda):  1 
    #> 
    #> Function to Estimate Density Ratio:
    #>   compute_density_ratio()

- **Kernel type** is fixed as Gaussian.
- **Number of kernels** is the number of kernels in the linear model.
  You can change by setting `kernel_num` argument. In default,
  `kernel_num = 100`.
- **Bandwidth (sigma)** is the Gaussian kernel bandwidth. In default,
  `sigma = "auto"`, the algorithm automatically select an optimal value
  by cross validation. If you set `sigma` a number, that will be used.
  If you set `sigma` a numeric vector, the algorithm select an optimal
  value in them by cross validation.
- **Centers** are centers of Gaussian kernels in the linear model. These
  are selected at random from the data sample `x1` underlying a
  numerator distribution `p(x)`. You can find the whole values in
  `result$kernel_info$centers`.
- **Kernel Weights** are `theta` parameters in the linear kernel model.
  You can find these values in `result$kernel_weights`.
- **Function to Estimate Density Ratio** is named
  `compute_density_ratio()`.

## 4. Multi Dimensional Data Samples

So far, the input data samples `x1` and `x2` were one dimensional.
`densratio()` allows to input multidimensional data samples as `matrix`,
as long as their dimensions are the same.

For example,

``` r
library(densratio)
library(mvtnorm)

set.seed(3)
x1 <- rmvnorm(300, mean = c(1, 1), sigma = diag(1/8, 2))
x2 <- rmvnorm(300, mean = c(1, 1), sigma = diag(1/2, 2))

densratio_obj_2d <- densratio(x1, x2)
densratio_obj_2d
#> 
#> Call:
#> densratio(x1 = x1, x2 = x2, method = "uLSIF")
#> 
#> Kernel Information:
#>   Kernel type:  Gaussian 
#>   Number of kernels:  100 
#>   Bandwidth (sigma):  0.316 
#>   Centers:  num [1:100, 1:2] 1.257 0.758 1.122 1.3 1.386 ...
#> 
#> Kernel Weights:
#>   num [1:100] 0.0756 0.0986 0.059 0.0797 0.0421 ...
#> 
#> Regularization Parameter (lambda):  0.3162278 
#> 
#> Function to Estimate Density Ratio:
#>   compute_density_ratio()
```

In this case, as well, we can compare the true density ratio with the
estimated density ratio.

``` r
true_density_ratio <- function(x) {
  dmvnorm(x, mean = c(1, 1), sigma = diag(1/8, 2)) /
    dmvnorm(x, mean = c(1, 1), sigma = diag(1/2, 2))
}

N <- 20
range <- seq(0, 2, length.out = N)
input <- expand.grid(range, range)
w_true <- matrix(true_density_ratio(input), nrow = N)
w_hat <- matrix(densratio_obj_2d$compute_density_ratio(input), nrow = N)

par(mfrow = c(1, 2))
contour(range, range, w_true, main = "True Density Ratio")
contour(range, range, w_hat, main = "Estimated Density Ratio")
```

<img src="man/figures/README-compare-2d-1.png" alt="" width="700" />

## 5. Used in research

The **densratio** package has been used in several research papers,
including:

- Fricke, K.R., & Herzberg, P.Y. (2019). Know your big data: De-biasing
  subsamples of large datasets for personality research using importance
  sampling and kNN matching.
- Chen, M.F., Goel, K., Sohoni, N.S., Poms, F., Fatahalian, K., &
  R’e, C. (2021). Mandoline: Model Evaluation under Distribution Shift.
  *International Conference on Machine Learning.*
- Evans, C., & G’Sell, M. (2024). Sequential label shift detection in
  classification data: An application to dengue fever. *PLOS ONE, 19.*
- Chakraborty, A., Huan, X., & Catanach, T.A. (2024). A Likelihood-Free
  Approach to Goal-Oriented Bayesian Optimal Experimental Design.
- Awaya, N., Xu, Y., & Ma, L. (2025). Two-sample comparison through
  additive tree models for density ratios.

## 6. Related work

- A Python Package for Density Ratio Estimation
  - <https://pypi.org/project/densratio/>
- APPEstimation: Adjusted Prediction Model Performance Estimation
  - <https://cran.r-project.org/package=APPEstimation>

## References

- Hido, S., Y. Tsuboi, H. Kashima, M. Sugiyama, and T. Kanamori.
  **Statistical outlier detection using direct density ratio
  estimation.** Knowledge and Information Systems, 2011.
- Kanamori, T., S. Hido, and M. Sugiyama. **A least-squares approach to
  direct importance estimation.** Journal of Machine Learning Research,
  2009.
- Liu, S., M. Yamada, N. Collier, M. Sugiyama. **Change-point detection
  in time-series data by relative density-ratio estimation.** Neural
  Net, 2013
- Sugiyama, M., S. Nakajima, H. Kashima, P. von Bünau, and M. Kawanabe.
  **Direct importance estimation with model selection and its
  application to covariate shift adaptation.** NIPS 2007.
- Sugiyama, M., T. Suzuki, and T. Kanamori. **Density ratio estimation
  in machine learning.** Cambridge University Press, 2012.
- Yamada, M., T. Suzuki, T. Kanamori, H. Hachiya, and M. Sugiyama.
  **Relative density-ratio estimation for robust distribution
  comparison.** NIPS 2011.

#' Estimate Density Ratio p(x)/q(x) by uLSIF (unconstrained Least-Square Importance Fitting)
#'
#' @param x1 numeric vector or matrix. Data from a numerator distribution p(x).
#' @param x2 numeric vector or matrix. Data from a denominator distribution q(x).
#' @param sigma positive numeric vector. Search range of Gaussian kernel bandwidth.
#' @param lambda positive numeric vector. Search range of regularization parameter.
#' @param kernel_num positive integer. Number of kernels.
#' @param verbose logical(default TRUE).
#'
#' @return uLSIF object that contains a function to compute estimated density ratio.
#'
#' @export
uLSIF <- function(x1, x2,
                  sigma = 10 ^ seq(-3, 1, length.out = 9),
                  lambda = 10 ^ seq(-3, 1, length.out = 9),
                  kernel_num = 100, verbose = TRUE) {

  params <- alist(x1 = x1, x2 = x2,
                  sigma = sigma, lambda = lambda, alpha = 0,
                  kernel_num = kernel_num, verbose = verbose)
  result <- do.call(RuLSIF, params)
  class(result) <- c("uLSIF", class(result))
  result
}

#' Estimate Density Ratio p(x)/q(x)
#'
#' @param x1 numeric vector or matrix. Data from a numerator distribution p(x).
#' @param x2 numeric vector or matrix. Data from a denominator distribution q(x).
#' @param method "uLSIF" (default), "RuLSIF", or "KLIEP".
#' @param sigma positive numeric vector. Search range of Gaussian kernel bandwidth.
#' @param lambda positive numeric vector. Search range of regularization parameter for uLSIF and RuLSIF.
#' @param alpha numeric in [0, 1]. Relative parameter for RuLSIF. Default 0.1.
#' @param kernel_num positive integer. Number of kernels.
#' @param fold positive integer. Numer of the folds of cross validation for KLIEP.
#' @param verbose logical (default TRUE).
#'
#' @return densratio object that contains a function to compute estimated density ratio.
#'
#' @examples
#' x1 <- rnorm(200, mean = 1, sd = 1/8)
#' x2 <- rnorm(200, mean = 1, sd = 1/2)
#'
#' densratio_obj <- densratio(x1, x2)
#'
#' new_x <- seq(0, 2, by = 0.05)
#' estimated_density_ratio <- densratio_obj$compute_density_ratio(new_x)
#'
#' plot(new_x, estimated_density_ratio, pch=19)
#'
#' @export
densratio <- function(x1, x2, method = c("uLSIF", "RuLSIF", "KLIEP"),
                      sigma = "auto", lambda = "auto", alpha = 0.1,
                      kernel_num = 100, fold = 5, verbose = TRUE) {
  # Prepare Arguments -------------------------------------------------------
  method <- match.arg(method)

  # To Retain Default Arguments in Functions of Methods ---------------------
  params <- alist(x1 = x1, x2 = x2, kernel_num = kernel_num, verbose = verbose)
  if (!identical(sigma, "auto")) {
    params <- c(params, alist(sigma = sigma))
  }

  # Run ---------------------------------------------------------------------
  if (method == "uLSIF") {
    if (!identical(lambda, "auto")) params <- c(params, alist(lambda = lambda))
    result <- do.call(uLSIF, params)
  } else if (method == "RuLSIF") {
    if (!identical(lambda, "auto")) params <- c(params, alist(lambda = lambda))
    params <- c(params, alist(alpha = alpha))
    result <- do.call(RuLSIF, params)
  } else {
    params <- c(params, alist(fold = fold))
    result <- do.call(KLIEP, params)
  }

  # Result ------------------------------------------------------------------
  cl <- match.call()
  cl[["method"]] <- method
  result$call <- cl

  class(result) <- c("densratio", class(result))
  result
}

KLIEP_search_sigma <- function(x, y, centers, fold, verbose = TRUE) {
  sigma <- 10
  score <- -Inf

  for (digit_position in 0:-5) {
    for (i in 1:9) {
      sigma_new <- sigma - 10 ^ digit_position

      score_new <- KLIEP_compute_score_cv(sigma_new, x, y, centers, fold)
      if (score_new <= score) {
        break
      }
      score <- score_new
      sigma <- sigma_new
      if(verbose) message(sprintf("  sigma = %.5f, score = %f", sigma, score))
    }
  }
  sigma
}

#' Estimate alpha-Relative Density Ratio p(x)/(alpha p(x) + (1-alpha) q(x))
#' by RuLSIF (Relative unconstrained Least-Square Importance Fitting)
#'
#' @param x1 numeric vector or matrix. Data from a numerator distribution p(x).
#' @param x2 numeric vector or matrix. Data from a denominator distribution q(x).
#' @param sigma positive numeric vector. Search range of Gaussian kernel bandwidth.
#' @param lambda positive numeric vector. Search range of regularization parameter.
#' @param alpha numeric value from 0.0 to 1.0. Relative parameter. Default 0.1.
#' @param kernel_num positive integer. Number of kernels.
#' @param verbose logical. Default TRUE.
#'
#' @return RuLSIF object which has `compute_density_ratio()`.
#'
#' @export
RuLSIF <- function(x1, x2,
                  sigma = 10 ^ seq(-3, 1, length.out = 9),
                  lambda = 10 ^ seq(-3, 1, length.out = 9),
                  alpha = 0.1, kernel_num = 100, verbose = TRUE) {

  if (verbose) message("################## Start RuLSIF ##################")
  if (is.vector(x1)) x1 <- matrix(x1)
  if (is.vector(x2)) x2 <- matrix(x2)
  if (ncol(x1) != ncol(x2)) stop("x1 and x2 must be same dimensions.")

  nx1 <- nrow(x1)
  nx2 <- nrow(x2)

  kernel_num <- min(kernel_num, nx1)
  centers <- x1[sample(nx1, size = kernel_num), , drop = FALSE]

  if(length(sigma) != 1 || length(lambda) != 1) {
    if(verbose) message("Searching optimal sigma and lambda...")
    opt_params <- RuLSIF_search_sigma_and_lambda(x1, x2, centers, sigma, lambda, alpha, verbose)
    sigma <- opt_params$sigma
    lambda <- opt_params$lambda
    if(verbose) message(sprintf("Found optimal sigma = %.3f, lambda = %.3f.", sigma, lambda))
  }

  if(verbose) message("Optimizing kernel weights...")
  phi_x1 <- compute_kernel_Gaussian(x1, centers, sigma)
  phi_x2 <- compute_kernel_Gaussian(x2, centers, sigma)
  H <- alpha * crossprod(phi_x1) / nx1 + (1 - alpha) * crossprod(phi_x2) / nx2
  h <- colMeans(phi_x1)
  kernel_weights <- solve(H + diag(lambda, kernel_num, kernel_num)) %*% h
  kernel_weights[kernel_weights < 0] <- 0
  if(verbose) message("End.")

  result <- list(kernel_weights = as.vector(kernel_weights),
                 lambda = lambda,
                 alpha = alpha,
                 kernel_info = list(
                   kernel = "Gaussian",
                   kernel_num = kernel_num,
                   sigma = sigma,
                   centers = centers
                 ),
                 compute_density_ratio = function(x) {
                   if(is.vector(x)) x <- matrix(x)
                   phi_x <- compute_kernel_Gaussian(x, centers, sigma)
                   density_ratio <- as.vector(phi_x %*% kernel_weights)
                   density_ratio
                 }
  )
  class(result) <- c("RuLSIF", class(result))
  if(verbose) message("################## Finished RuLSIF ###############")
  result
}

#' @importFrom utils str
#' @export
print.densratio <- function(x, digits = 3L, ...) {
  if (!is.null(x$call)) {
    cat("\nCall:\n", paste(deparse(x$call), sep = "\n", collapse = "\n"),
        "\n\n", sep = "")
  }

  info <- x$kernel_info
  cat("Kernel Information:\n")
  cat("  Kernel type: ", info$kernel, "\n")
  cat("  Number of kernels: ", info$kernel_num, "\n")
  cat("  Bandwidth (sigma): ", format(info$sigma, digits = digits), "\n")
  cat("  Centers: ")
  str(info$centers, digits.d = digits, give.attr = FALSE)
  cat("\n")
  cat("Kernel Weights:\n ")
  str(x$kernel_weights, digits.d = digits, give.attr = FALSE)
  cat("\n")

  if ("uLSIF" %in% class(x)) {
    cat("Regularization Parameter (lambda): ", x$lambda, "\n\n")
  } else if ("RuLSIF" %in% class(x)) {
    cat("Regularization Parameter (lambda): ", x$lambda, "\n")
    cat("Relative Parameter (alpha): ", x$alpha, "\n\n")
  } else if ("KLIEP" %in% class(x)) {
    cat("Number of the Folds: ", x$fold, "\n\n")
  }

  cat("Function to Estimate Density Ratio:\n")
  cat("  compute_density_ratio()\n")

  cat("\n")
  invisible(x)
}

#' @export
print.KLIEP <- print.densratio

#' @export
print.uLSIF <- print.densratio

#' @export
print.RuLSIF <- print.densratio

KLIEP_optimize_alpha <- function(phi_x, phi_y, mean_phi_y) {
  A <- phi_x
  if(missing(phi_y)) {
    if(is.matrix(mean_phi_y)) {
      b <- mean_phi_y
    } else {
      b <- matrix(mean_phi_y)
    }
  } else {
    b <- matrix(colMeans(phi_y))
  }
  c <- b / crossprod(b)[1,1]

  kernel_num <- ncol(phi_x)
  max_iteration <- 100

  alpha <- matrix(rep(1, kernel_num))
  alpha <- compute_next_alpha(alpha, b, c)
  score <- KLIEP_compute_score(phi_x, alpha)

  eplilon_list <- 10 ^ (3:-3)
  for (epsilon in eplilon_list) {
    for (i in seq_len(max_iteration)) {
      alpha_new <- alpha + (epsilon * t(A)) %*% (1 / (A %*% alpha))
      alpha_new <- compute_next_alpha(alpha_new, b, c)
      score_new <- KLIEP_compute_score(phi_x, alpha_new)
      if(score_new <= score) {
        break
      }
      alpha <- alpha_new
      score <- score_new
    }
  }
  alpha
}

compute_next_alpha <- function(alpha, b, c = b / crossprod(b)[1,1]) {
  alpha <- alpha + (1 - (t(b) %*% alpha)[1,1]) * c
  alpha[alpha < 0, ] <- 0
  alpha <- alpha / (t(b) %*% alpha)[1,1]
  alpha
}

KLIEP_compute_score_cv <- function(sigma, x, y, centers, fold) {
  phi_x <- compute_kernel_Gaussian(x, centers, sigma)
  phi_y <- compute_kernel_Gaussian(y, centers, sigma)
  mean_phi_y <- matrix(colMeans(phi_y))

  nx <- nrow(x)
  cv_split <- sample(nx) %% fold + 1

  scores <- numeric(fold)
  for (i in seq_len(fold)) {
    alpha <- KLIEP_optimize_alpha(phi_x[cv_split != i, ], mean_phi_y = mean_phi_y)
    scores[i] <- KLIEP_compute_score(phi_x[cv_split == i, ], alpha)
  }
  score_new <- mean(scores)
  score_new
}

KLIEP_compute_score <- function(phi, alpha) {
  mean(log(phi %*% alpha))
}

compute_kernel_Gaussian <- function(x, centers, sigma) {
  apply(centers, 1, function(center) {
    apply(x, 1, function(row) {
      kernel_Gaussian(row, center, sigma)
    })
  })
}

kernel_Gaussian <- function(x, y, sigma) {
  exp(- squared_euclid_distance(x, y) / (2 * sigma * sigma))
}

#' Compute Squared Euclid Distance
#'
#' @param x a numeric vector.
#' @param y a numeric vector.
#'
#' @return squared Euclid distance
squared_euclid_distance <- function(x, y) {
  sum((x - y) ^ 2)
}

test_that("squared_euclid_distance computes squared Euclidean distance", {
  expect_equal(squared_euclid_distance(c(1, 2, 3), c(4, 6, 3)), 25)
})

test_that("kernel_Gaussian is 1 for identical inputs", {
  expect_equal(kernel_Gaussian(c(1, 2), c(1, 2), sigma = 1), 1)
})

test_that("compute_kernel_Gaussian returns a kernel matrix", {
  x <- matrix(c(0, 0, 1, 0, 0, 1), ncol = 2, byrow = TRUE)
  centers <- matrix(c(0, 0, 1, 1), ncol = 2, byrow = TRUE)

  result <- compute_kernel_Gaussian(x, centers, sigma = 1)

  expect_equal(dim(result), c(3, 2))
  expect_true(all(result > 0))
  expect_true(all(result <= 1))
})

RuLSIF_search_sigma_and_lambda <- function(x, y, centers, sigma_list, lambda_list, alpha, verbose) {
  nx <- nrow(x)
  ny <- nrow(y)
  n_min <- min(nx, ny)
  kernel_num <- nrow(centers)

  score_new <- Inf
  sigma_new <- 0
  lambda_new <- 0
  for (sigma in sigma_list) {
    phi_x <- compute_kernel_Gaussian(x, centers, sigma)
    phi_y <- compute_kernel_Gaussian(y, centers, sigma)
    H <- alpha * crossprod(phi_x) / nx + (1 - alpha) * crossprod(phi_y) / ny
    h <- colMeans(phi_x)
    phi_x <- t(phi_x[seq_len(n_min), , drop = FALSE])
    phi_y <- t(phi_y[seq_len(n_min), , drop = FALSE])
    for (lambda in lambda_list) {
      B <- H + diag(lambda * (ny - 1) / ny, nrow = kernel_num, ncol = kernel_num)
      B_inv <- solve(B)
      B_inv_X <- B_inv %*% phi_y
      X_B_inv_X <- phi_y * B_inv_X
      denom <- ones(n_min, value = ny) - ones(kernel_num) %*% X_B_inv_X
      B0 <- B_inv %*% (h %*% ones(n_min)) +
        B_inv_X %*% diag(as.vector((t(h) %*% B_inv_X) / denom))
      B1 <- B_inv %*% phi_x +
        B_inv_X %*% diag(as.vector((ones(kernel_num) %*% (phi_x * B_inv_X)) / denom))
      B2 <- (ny-1) * (nx * B0 - B1) / (ny * (nx - 1))
      B2[B2 < 0] <- 0
      r_y <- t(ones(kernel_num) %*% (phi_y * B2))
      r_x <- t(ones(kernel_num) %*% (phi_x * B2))
      score <- (crossprod(r_y) / 2 - ones(n_min) %*% r_x) / n_min
      if(score < score_new) {
        if(verbose) message(sprintf("  sigma = %.3f, lambda = %.3f, score = %.3f", sigma, lambda, score))
        score_new <- score
        sigma_new <- sigma
        lambda_new <- lambda
      }
    }
  }
  list(sigma = sigma_new, lambda = lambda_new)
}

ones <- function(size, value=1) {
  t(rep(value, size))
}

KLIEP_search_sigma_list <- function(x, y, centers, sigma_list, fold, verbose = TRUE) {
  sigma <- NULL
  score <- -Inf

  for(sigma_new in sigma_list) {
    score_new <- KLIEP_compute_score_cv(sigma_new, x, y, centers, fold)
    if (score_new > score) {
      score <- score_new
      sigma <- sigma_new
      if(verbose) message(sprintf("  sigma = %.5f, score = %f", sigma, score))
    }
  }
  sigma
}

#' Estimate Density Ratio p(x)/q(x) by KLIEP (Kullback-Leibler Importance Estimation Procedure)
#'
#' @param x1 numeric vector or matrix. Data from a numerator distribution p(x).
#' @param x2 numeric vector or matrix. Data from a denominator distribution q(x).
#' @param sigma positive numeric vector. Search range of Gaussian kernel bandwidth.
#' @param kernel_num positive integer. Number of kernels.
#' @param fold positive integer. Number of the folds of cross validation.
#' @param verbose logical (default TRUE).
#'
#' @return KLIEP object that contains a function to compute estimated density ratio.
#'
#' @export
KLIEP <- function(x1, x2, sigma = "auto", kernel_num = 100, fold = 5, verbose = TRUE) {
  if(verbose) message("############ Start KLIEP ############")
  if(is.vector(x1)) x1 <- matrix(x1)
  if(is.vector(x2)) x2 <- matrix(x2)
  if(ncol(x1) != ncol(x2)) stop("x1 and x2 must be same dimensions.")

  nx1 <- nrow(x1)
  kernel_num <- min(kernel_num, nx1)
  centers <- x1[sample(nx1, size = kernel_num), , drop = FALSE]

  if(identical(sigma, "auto")) {
    if(verbose) message("Searching optimal sigma and lambda...")
    sigma <- KLIEP_search_sigma(x1, x2, centers, fold, verbose)
    if(verbose) message(sprintf("Found optimal sigma = %.5f.", sigma))
  } else if(length(sigma) > 1) {
    if(verbose) message("Searching optimal sigma and lambda...")
    sigma <- KLIEP_search_sigma_list(x1, x2, centers, sigma, fold, verbose)
    if(verbose) message(sprintf("Found optimal sigma = %.5f.", sigma))
  }

  if(verbose) message("Optimizing kernel weights...")
  phi_x1 <- compute_kernel_Gaussian(x1, centers, sigma)
  phi_x2 <- compute_kernel_Gaussian(x2, centers, sigma)
  kernel_weights <- KLIEP_optimize_alpha(phi_x1, phi_x2)
  if(verbose) message("End.")

  result <- list(kernel_weights = as.vector(kernel_weights),
                 kernel_info = list(
                   kernel = "Gaussian",
                   kernel_num = kernel_num,
                   sigma = sigma,
                   centers = centers
                 ),
                 fold = fold,
                 compute_density_ratio = function(x) {
                   if (is.vector(x)) x <- matrix(x)
                   phi_x <- compute_kernel_Gaussian(x, centers, sigma)
                   density_ratio <- as.vector(phi_x %*% kernel_weights)
                   density_ratio
                 }
  )
  class(result) <- c("KLIEP", class(result))
  if(verbose) message("############ Finished KLIEP #########")
  result
}

context("KLIEP")

test_that("KLIEP", {
  set.seed(314)
  x <- rnorm(200, mean = 1, sd = 1/8)
  y <- rnorm(200, mean = 1, sd = 1/2)

  result <- KLIEP(x, y, sigma = 0.1, kernel_num = 20, fold = 5, verbose = FALSE)

  expect_true(inherits(result, "KLIEP"))
  expect_equal(result$kernel_info$kernel, "Gaussian")
  expect_equal(result$kernel_info$kernel_num, 20)
  expect_equal(result$kernel_info$sigma, 0.1)
  expect_equal(result$fold, 5)

  expect_equal(length(result$kernel_weights), 20)
  expect_true(all(is.finite(result$kernel_weights)))
  expect_true(all(result$kernel_weights >= 0))
  expect_true(is.function(result$compute_density_ratio))
})

test_that("KLIEP compute_density_ratio uses new input", {
  set.seed(314)
  x <- rnorm(200, mean = 1, sd = 1 / 8)
  y <- rnorm(200, mean = 1, sd = 1 / 2)

  result <- KLIEP(x, y, sigma = 0.1, kernel_num = 20, fold = 5, verbose = FALSE)

  new_x <- seq(0, 2, length.out = 11)
  density_ratio <- result$compute_density_ratio(new_x)

  expect_length(density_ratio, 11)
  expect_true(all(is.finite(density_ratio)))
  expect_true(all(density_ratio >= 0))
})

---
title: "An R Package for Density Ratio Estimation"
author: "Koji MAKIYAMA (@hoxo-m)"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{An R Package for Density Ratio Estimation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, message = FALSE)
library(mvtnorm)
```

## 1. Overview

**Density ratio estimation** is described as follows: 
for given two data samples $x1$ and $x2$ from unknown distributions $p(x)$ and $q(x)$ respectively, estimate

$$
w(x) = \frac{p(x)}{q(x)}
$$

where $x1$ and $x2$ are $d$-dimensional real numbers.

The estimated density ratio function $w(x)$ can be used in many applications such as **anomaly detection** [Hido et al. 2011], **change-point detection** [Liu et al. 2013], and **covariate shift adaptation** [Sugiyama et al. 2007].
Other useful applications about density ratio estimation were summarized by [Sugiyama et al. 2012].

The package **densratio** provides a function `densratio()` that returns an object with a method to estimate density ratio as `compute_density_ratio()`.

For example, 

```{r}
set.seed(3)
x1 <- rnorm(200, mean = 1, sd = 1/8)
x2 <- rnorm(200, mean = 1, sd = 1/2)

library(densratio)
densratio_obj <- densratio(x1, x2)
```

The function `densratio()` estimates the density ratio of $p(x)$ to $q(x)$,
$$
w(x) = \frac{p(x)}{q(x)} = \frac{\rm{Norm}(1, 1/8)}{\rm{Norm}(1, 1/2)}
$$
and provides a function to compute estimated density ratio.

The densratio object has a function `compute_density_ratio()` that can compute density ratio $\hat{w}(x) \simeq p(x)/q(x)$ for any $d$-dimensional input $x$ (here $d=1$).

```{r fig.width=5, fig.height=4}
new_x <- seq(0, 2, by = 0.05)
w_hat <- densratio_obj$compute_density_ratio(new_x)

plot(new_x, w_hat, pch=19)
```

In this case, the true density ratio $w(x) = p(x)/q(x) = \rm{Norm}(1, 1/8) / \rm{Norm}(1, 1/2)$ is known.
So we can compare $w(x)$ with the estimated density ratio $\hat{w}(x)$.

```{r fig.width=5, fig.height=4}
true_density_ratio <- function(x) dnorm(x, 1, 1/8) / dnorm(x, 1, 1/2)

plot(true_density_ratio, xlim=c(0, 2), lwd=2, col="red", xlab = "x", ylab = "Density Ratio")
plot(densratio_obj$compute_density_ratio, xlim=c(0, 2), lwd=2, col="green", add=TRUE)
legend("topright", legend=c(expression(w(x)), expression(hat(w)(x))), col=2:3, lty=1, lwd=2, pch=NA)
```

## 2. How to Install

You can install the **densratio** package from [CRAN](https://CRAN.R-project.org/package=densratio).

```{r eval=FALSE}
install.packages("densratio")
```

You can also install the package from [GitHub](https://github.com/hoxo-m/densratio).

```{r eval=FALSE}
install.packages("remotes") # If you have not installed "remotes" package
remotes::install_github("hoxo-m/densratio")
```

The source code for **densratio** package is available on GitHub at

- https://github.com/hoxo-m/densratio.

## 3. Details

### 3.1. Basics

The package provides `densratio()`.
The function returns an object that has a function to compute estimated density ratio.

For data samples `x1` and `x2`,

```{r eval=FALSE}
library(densratio)

x1 <- rnorm(200, mean = 1, sd = 1/8)
x2 <- rnorm(200, mean = 1, sd = 1/2)

result <- densratio(x1, x2)
```

In this case, `densratio_obj$compute_density_ratio()` can compute estimated density ratio.

```{r fig.width=5, fig.height=4}
new_x <- seq(0, 2, by = 0.05)
w_hat <- densratio_obj$compute_density_ratio(new_x)

plot(new_x, w_hat, pch=19)
```

### 3.2. Methods

`densratio()` has `method` argument that you can pass `"uLSIF"`, `"RuSLIF"`, or `"KLIEP"`.

- **uLSIF** (unconstrained Least-Squares Importance Fitting) is the default method.
This algorithm estimates density ratio by minimizing the squared loss.
You can find more information in [Kanamori et al. 2009] and [Hido et al. 2011].
- **RuLSIF** (Relative unconstrained Least-Squares Importance Fitting).
This algorithm estimates relative density ratio by minimizing the squared loss.
You can find more information in [Yamada et al. 2011] and [Liu et al. 2013].
- **KLIEP** (Kullback-Leibler Importance Estimation Procedure).
This algorithm estimates density ratio by minimizing Kullback-Leibler divergence.
You can find more information in [Sugiyama et al. 2007].

The methods assume that density ratio are represented by linear model:

$$
w(x) = \theta_1 K(x, c_1) + \theta_2 K(x, c_2) + ... + \theta_b K(x, c_b)
$$

where

$$
K(x, c) = \exp\left(-\frac{\|x - c\|^2}{2 \sigma ^ 2}\right)
$$

is the Gaussian (RBF) kernel.

`densratio()` performs the following: 

- Decides kernel parameter $\sigma$ by cross-validation, 
- Optimizes the kernel weights $\theta$ (in other words, find the optimal coefficients of the linear model), and
- The parameters $\sigma$ and $\theta$ are saved into `densratio` object, and are used when to compute density ratio in the call `compute_density_ratio()`.

### 3.3. Result and Arguments

You can display information of densratio objects.
Moreover, you can change some conditions to specify arguments of `densratio()`.

```{r}
densratio_obj
```

- **Kernel type** is fixed as Gaussian.
- **Number of kernels** is the number of kernels in the linear model. 
You can change by setting `kernel_num` argument. 
In default, `kernel_num = 100`.
- **Bandwidth (sigma)** is the Gaussian kernel bandwidth.
In default, `sigma = "auto"`, the algorithm automatically select an optimal value by cross validation. 
If you set `sigma` a number, that will be used. 
If you set `sigma` a numeric vector, the algorithm select an optimal value in them by cross validation.
- **Centers** are centers of Gaussian kernels in the linear model. 
These are selected at random from the data sample `x1` underlying a numerator distribution $p(x)$. 
You can find the whole values in `result$kernel_info$centers`.
- **Kernel Weights** are `theta` parameters in the linear kernel model. 
You can find these values in `result$kernel_weights`.
- **Function to Estimate Density Ratio** is named `compute_density_ratio()`.

## 4. Multi Dimensional Data Samples

So far, the input data samples `x1` and `x2` were one dimensional. 
`densratio()` allows to input multidimensional data samples as `matrix`, as long as their dimensions are the same.

For example,

```{r}
library(densratio)
library(mvtnorm)

set.seed(3)
x1 <- rmvnorm(300, mean = c(1, 1), sigma = diag(1/8, 2))
x2 <- rmvnorm(300, mean = c(1, 1), sigma = diag(1/2, 2))

densratio_obj_d2 <- densratio(x1, x2)
densratio_obj_d2
```

In this case, as well, we can compare the true density ratio with the estimated density ratio.

```{r fig.width=7, fig.height=4}
true_density_ratio <- function(x) {
  dmvnorm(x, mean = c(1, 1), sigma = diag(1/8, 2)) /
    dmvnorm(x, mean = c(1, 1), sigma = diag(1/2, 2))
}

N <- 20
range <- seq(0, 2, length.out = N)
input <- expand.grid(range, range)
w_true <- matrix(true_density_ratio(input), nrow = N)
w_hat <- matrix(densratio_obj_d2$compute_density_ratio(input), nrow = N)

par(mfrow = c(1, 2))
contour(range, range, w_true, main = "True Density Ratio")
contour(range, range, w_hat, main = "Estimated Density Ratio")
```

## 5. Related work

- A Python Package for Density Ratio Estimation
    - https://pypi.org/project/densratio/
- APPEstimation: Adjusted Prediction Model Performance Estimation
    - https://cran.r-project.org/package=APPEstimation

## References

- Hido, S., Y. Tsuboi, H. Kashima, M. Sugiyama, and T. Kanamori.
**Statistical outlier detection using direct density ratio estimation.**
Knowledge and Information Systems, 2011.
- Kanamori, T., S. Hido, and M. Sugiyama.
**A least-squares approach to direct importance estimation.**
Journal of Machine Learning Research, 2009.
- Liu, S., M. Yamada, N. Collier, M. Sugiyama.
**Change-point detection in time-series data by relative density-ratio estimation.**
Neural Net, 2013
- Sugiyama, M., S. Nakajima, H. Kashima, P. von Bünau, and M. Kawanabe. 
**Direct importance estimation with model selection and its application to covariate shift adaptation.**
NIPS 2007.
- Sugiyama, M., T. Suzuki, and T. Kanamori. 
**Density ratio estimation in machine learning.**
Cambridge University Press, 2012.
- Yamada, M., T. Suzuki, T. Kanamori, H. Hachiya, and M. Sugiyama.
**Relative density-ratio estimation for robust distribution comparison.**
NIPS 2011.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RuLSIF.R
\name{RuLSIF}
\alias{RuLSIF}
\title{Estimate alpha-Relative Density Ratio p(x)/(alpha p(x) + (1-alpha) q(x))
by RuLSIF (Relative unconstrained Least-Square Importance Fitting)}
\usage{
RuLSIF(
  x1,
  x2,
  sigma = 10^seq(-3, 1, length.out = 9),
  lambda = 10^seq(-3, 1, length.out = 9),
  alpha = 0.1,
  kernel_num = 100,
  verbose = TRUE
)
}
\arguments{
\item{x1}{numeric vector or matrix. Data from a numerator distribution p(x).}

\item{x2}{numeric vector or matrix. Data from a denominator distribution q(x).}

\item{sigma}{positive numeric vector. Search range of Gaussian kernel bandwidth.}

\item{lambda}{positive numeric vector. Search range of regularization parameter.}

\item{alpha}{numeric value from 0.0 to 1.0. Relative parameter. Default 0.1.}

\item{kernel_num}{positive integer. Number of kernels.}

\item{verbose}{logical. Default TRUE.}
}
\value{
RuLSIF object which has `compute_density_ratio()`.
}
\description{
Estimate alpha-Relative Density Ratio p(x)/(alpha p(x) + (1-alpha) q(x))
by RuLSIF (Relative unconstrained Least-Square Importance Fitting)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/densratio.R
\name{densratio}
\alias{densratio}
\title{Estimate Density Ratio p(x)/q(x)}
\usage{
densratio(
  x1,
  x2,
  method = c("uLSIF", "RuLSIF", "KLIEP"),
  sigma = "auto",
  lambda = "auto",
  alpha = 0.1,
  kernel_num = 100,
  fold = 5,
  verbose = TRUE
)
}
\arguments{
\item{x1}{numeric vector or matrix. Data from a numerator distribution p(x).}

\item{x2}{numeric vector or matrix. Data from a denominator distribution q(x).}

\item{method}{"uLSIF" (default), "RuLSIF", or "KLIEP".}

\item{sigma}{positive numeric vector. Search range of Gaussian kernel bandwidth.}

\item{lambda}{positive numeric vector. Search range of regularization parameter for uLSIF and RuLSIF.}

\item{alpha}{numeric in [0, 1]. Relative parameter for RuLSIF. Default 0.1.}

\item{kernel_num}{positive integer. Number of kernels.}

\item{fold}{positive integer. Numer of the folds of cross validation for KLIEP.}

\item{verbose}{logical (default TRUE).}
}
\value{
densratio object that contains a function to compute estimated density ratio.
}
\description{
Estimate Density Ratio p(x)/q(x)
}
\examples{
x1 <- rnorm(200, mean = 1, sd = 1/8)
x2 <- rnorm(200, mean = 1, sd = 1/2)

densratio_obj <- densratio(x1, x2)

new_x <- seq(0, 2, by = 0.05)
estimated_density_ratio <- densratio_obj$compute_density_ratio(new_x)

plot(new_x, estimated_density_ratio, pch=19)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/uLSIF.R
\name{uLSIF}
\alias{uLSIF}
\title{Estimate Density Ratio p(x)/q(x) by uLSIF (unconstrained Least-Square Importance Fitting)}
\usage{
uLSIF(
  x1,
  x2,
  sigma = 10^seq(-3, 1, length.out = 9),
  lambda = 10^seq(-3, 1, length.out = 9),
  kernel_num = 100,
  verbose = TRUE
)
}
\arguments{
\item{x1}{numeric vector or matrix. Data from a numerator distribution p(x).}

\item{x2}{numeric vector or matrix. Data from a denominator distribution q(x).}

\item{sigma}{positive numeric vector. Search range of Gaussian kernel bandwidth.}

\item{lambda}{positive numeric vector. Search range of regularization parameter.}

\item{kernel_num}{positive integer. Number of kernels.}

\item{verbose}{logical(default TRUE).}
}
\value{
uLSIF object that contains a function to compute estimated density ratio.
}
\description{
Estimate Density Ratio p(x)/q(x) by uLSIF (unconstrained Least-Square Importance Fitting)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KLIEP.R
\name{KLIEP}
\alias{KLIEP}
\title{Estimate Density Ratio p(x)/q(x) by KLIEP (Kullback-Leibler Importance Estimation Procedure)}
\usage{
KLIEP(x1, x2, sigma = "auto", kernel_num = 100, fold = 5, verbose = TRUE)
}
\arguments{
\item{x1}{numeric vector or matrix. Data from a numerator distribution p(x).}

\item{x2}{numeric vector or matrix. Data from a denominator distribution q(x).}

\item{sigma}{positive numeric vector. Search range of Gaussian kernel bandwidth.}

\item{kernel_num}{positive integer. Number of kernels.}

\item{fold}{positive integer. Number of the folds of cross validation.}

\item{verbose}{logical (default TRUE).}
}
\value{
KLIEP object that contains a function to compute estimated density ratio.
}
\description{
Estimate Density Ratio p(x)/q(x) by KLIEP (Kullback-Leibler Importance Estimation Procedure)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_kernel_Gaussian.R
\name{squared_euclid_distance}
\alias{squared_euclid_distance}
\title{Compute Squared Euclid Distance}
\usage{
squared_euclid_distance(x, y)
}
\arguments{
\item{x}{a numeric vector.}

\item{y}{a numeric vector.}
}
\value{
squared Euclid distance
}
\description{
Compute Squared Euclid Distance
}
